import collections
import datetime
import resource
import sys
import pickle
import threading
import logging 
import traceback

log = logging.getLogger("domain_stats")

class cache_stats:
    def __init__(self,hit=0, miss=0, expire=0):
        self.hit = hit
        self.miss = miss
        self.expire = expire

    def __repr__(self):
        return f"cache_stats(hit={self.hit}, miss={self.miss}, expire={self.expire})"

    def reset(self):
        self.hit = self.miss = self.expire = 0


class ExpiringCache(collections.OrderedDict, collections.Counter):
    """This is a Least Recently Used Expiring Cache. Reading or setting a record makes it the most recently used.
       When an item is added such that the maxsize is exceeded the least recently used entry is dropped.
       Additionally entries will be marked with an expiration date.  If the expiration is exceeded None is retreieved
       and the entry is deleted when you query a value in the dictionary.""" 

    def __init__(self, maxsize = 65535, default_hours_to_live=720, *args, **kwargs):
        self.maxsize = maxsize
        #Set to -1 for no expirations and it behaves like a regular LRU cache 
        self.hours_to_live = default_hours_to_live
        self.stats = cache_stats()
        self.update_lock = threading.Lock()
        super().__init__()

    def __contains__(self,*args,**kwargs):
        try:
            if  sys._getframe().f_back.f_back.f_code != "get":
                print(f"Warning: __contains__ called with 'in' keyword.  This will likely skew your cache.stats.miss accuracy.  Instead just get() check for none being returned.")
                print( sys._getframe().f_back.f_code)
        except:
            pass
        return super().__contains__(*args,**kwargs)

    def cache_info(self):
        """JSON transmitable Report cache performance statistics"""
        rpt =  f"""{self.stats}, ('Max Size': {self.maxsize}, 'Current size': {len(self)}, 'Cache Bytes':{sys.getsizeof(self)}, 'Application Kilobytes':{resource.getrusage(resource.RUSAGE_SELF).ru_maxrss})"""
        return rpt

    def cache_report(self):
        """Detailed info about the cache for local consumption only"""
        rpt = "Cache Report\n{0:-^80}  {1:-^9}  {2:-^26}  {3}\n".format("KEY","COUNT","EXPIRES","RECORD---->")
        for thekey,rec in sorted(self.items(),key = lambda entry:entry[1][1], reverse=True):
            exp, cnt, entry = rec
            rpt += "{0: ^80}  {1:0>9}  {2: ^20}  {3}\n".format(str(thekey),str(cnt),str(exp),str(entry))
        log.info(rpt)
        return rpt

    def cache_dump(self, fname):
        log.debug(f"Dumping cache to file {fname}")
        with open(fname,"wb") as fhandle:
            pickle.dump(list(self.items()), fhandle, protocol=pickle.HIGHEST_PROTOCOL)

    def cache_load(self, fname):
        log.debug(f"Loading cache from file {fname}")
        self.clear()
        with open(fname, "rb") as fhandle:
            other = pickle.load(fhandle)
            for key,val in other:
                with self.update_lock:
                    super().__setitem__(key, val)

    def get(self,key,default_value=None):
        retval = self[key]
        if retval == None:
            return default_value
        return retval


    def __getitem__(self, key):
        if key in self:
            expiration, read_count, data = super().__getitem__(key)
            del self[key]
            #If it set to never expire or it is not expired then update the hit count and recreate the record.
            if (type(expiration)==int and expiration<0) or (type(expiration)==datetime.datetime and (expiration > datetime.datetime.utcnow())):
                self.stats.hit += 1
                with self.update_lock:
                    super().__setitem__(key, (expiration, read_count+1, data))
                return data
            else:
                self.stats.expire += 1
        else:
            self.stats.miss += 1
        return None

    def enforce_size(self):
        """Delete items from front of dict (First in) unless it has an expiration of -2 until the length is < self.maxsize"""
        entries = iter(self.keys())
        with self.update_lock:
            while len(self) > self.maxsize:
                try:
                    tgt_key= next(entries)
                except StopIteration:
                    print("Unable to delete any keys but the maximum size is exceeded.  Ignoring Maxsize.")
                    break
                else:
                    expires,_,_ = super().get(tgt_key)
                    if expires != -2:
                        del self[tgt_key]
                        entries = iter(self.keys())


    def set(self, key, value, update_expiration=True, reset_read_count=False, hours_to_live=None):
        """If update_expiration is True then the expiration date is updated on overwrites to existing keys"""
        """If you update an expired entry then a new expiration date is set on a new entry"""
        """If reset_read_count is True then read_count is set to sero on overwrites to existing  keys"""
        """Setting hours_to_live to -1 and they won't expire but can page out if least recently used."""
        """Set to -2 and they do not expire and the LRU can not remove them. Permanent entries in the cache (use with caution)"""
        """Set to 0 and the record is treated as already expired. nothing is added to the cache"""
        log.debug(f"cache set called {key} {value} {hours_to_live}")
        if hours_to_live == None:
            hours_to_live = self.hours_to_live
        if hours_to_live == 0:
            log.debug(f"hours to live set to zero.  Not caching. {key} {value}")
            return
        if hours_to_live < 0:
            expires = hours_to_live
        else:
            expires = datetime.datetime.utcnow() + datetime.timedelta(hours=hours_to_live)
        read_count = 0
        if key in self:
            with self.update_lock:
                current_expires,current_read_count,value = super().__getitem__(key)
                del self[key]
            if not update_expiration and (current_expires < datetime.datetime.utcnow()):
                expires = current_expires
            if not reset_read_count:
                read_count = current_read_count
        with self.update_lock:
            super().__setitem__( key, (expires, read_count, value))
        self.enforce_size()

    def __setitem__(self, key, value):
        if self.hours_to_live == 0:
            return
        if self.hours_to_live < 0:
            expires = self.hours_to_live
        else:
            expires = datetime.datetime.utcnow() + datetime.timedelta(hours=self.hours_to_live)
        read_count = 0
        if key in self:
            current_expires,read_count,value = super().__getitem__(key)
            del self[key]
            if current_expires < datetime.datetime.utcnow():
                expires = current_expires
        with self.update_lock:
            super().__setitem__( key, (expires, read_count, value))
        self.enforce_size()


def expiring_cache(maxsize=65535, cacheable = lambda _:True, hours_to_live=720):
    #Create my own lru cache so I can remove items as needed
    def wrap_function_with_cache(function_to_call):
        _cache =  ExpiringCache(hours_to_live=hours_to_live, cacheable=cacheable, maxsize=maxsize)
        def newfunc(*args):
            nonlocal _cache
            data = _cache[args]
            if data != None:
                return data
            ret_val = function_to_call(*args)
            #check to see if this should be cached
            if cacheable(ret_val):
                _cache[args] =  ret_val
            return ret_val

        def bypass_cache(*args):
            ret_val = function_to_call(*args)
            return ret_val

        newfunc.cache = _cache
        newfunc.reset_stats = _cache.stats.reset
        newfunc.cache_dump = _cache.cache_dump
        newfunc.cache_load = _cache.cache_load
        newfunc.cache_info = _cache.cache_info
        newfunc.bypass_cache = bypass_cache
        return newfunc
    return wrap_function_with_cache


import sqlite3
import threading
import datetime
import logging
import pathlib
import urllib
import os

log = logging.getLogger("domain_stats")

  
class database_stats:
    def __init__(self, hit=0,miss=0, insert=0, delete=0) :
        self.hit = hit
        self.miss = miss
        self.insert = insert
        self.delete = delete

    def __repr__(self):
        return f"database_stats(hit={self.hit},miss={self.miss},insert={self.insert},delete={self.delete})"

def reduce_domain(domain_in):
    parts =  domain_in.strip().split(".")
    if len(parts)> 2: 
        if parts[-1] not in ['com','org','net','gov','edu']:
            if parts[-2] in ['co', 'com','ne','net','or','org','go','gov','ed','edu','ac','ad','gr','lg','mus','gouv']:
                domain = ".".join(parts[-3:])
            else:
                domain = ".".join(parts[-2:])
        else:
            domain = ".".join(parts[-2:])
            #print("trim top part", domain_in, domain)
    else:
        domain = ".".join(parts)
    return domain.lower()

class DomainStatsDatabase(object):

    def __init__(self, filename):
        self.filename = filename
        self.lock = threading.Lock()
        self.stats = database_stats()
        if not pathlib.Path(self.filename).exists():
            print(f"WARNING: Database not found. {self.filename}")
            return
        db = sqlite3.connect(filename, timeout=15)
        cursor = db.cursor()
        self.version,self.created,self.lastupdate = cursor.execute("select version,created,lastupdate from info").fetchone()

    def create_file(self, filename):
        datab = sqlite3.connect(filename)
        cursor = datab.cursor()
        cursor.execute("CREATE TABLE domains (domain text NOT NULL UNIQUE,seen_by_web timestamp not NULL,expires timestamp not NULL,seen_by_isc timestamp not NULL,seen_by_you timestamp not NULL)")
        cursor.execute("CREATE TABLE info (version real NOT NULL,created timestamp not NULL,lastupdate timestamp not NULL)")
        new_info = (1.0, datetime.datetime.utcnow(), datetime.datetime.utcnow())
        cursor.execute("insert into info (version, created, lastupdate) values (?,?,?)", new_info)
        datab.commit()
        self.version, self.created, self.last_update = new_info

    def reset_first_contact(self):
        db= sqlite3.connect(self.filename)
        log.info("Database_admin was used to reset all of the seen_by_you dates to FIRST-CONTACT")
        print(f"Resetting all seen_by_you dates to FIRST-CONTACT.")
        cursor = db.cursor()
        cursor.execute("update domains set seen_by_you=?", ("FIRST-CONTACT",))
        db.commit() 
        print(f"RESET! You probably want to also delete your .cache file at this time.")

    def update_record(self, domain, record_seen_by_web, record_expires, record_seen_by_isc, record_seen_by_you):
        record_seen_by_web = record_seen_by_web.strftime('%Y-%m-%d %H:%M:%S')
        record_expires = record_expires.strftime('%Y-%m-%d %H:%M:%S')
        if record_seen_by_isc != "LOCAL" and record_seen_by_isc != "RDAP":
            record_seen_by_isc = record_seen_by_isc.strftime('%Y-%m-%d %H:%M:%S')
        if record_seen_by_you != "FIRST-CONTACT":
            record_seen_by_you = record_seen_by_you.strftime('%Y-%m-%d %H:%M:%S')
        db = sqlite3.connect(self.filename, timeout=15)
        cursor = db.cursor()
        log.info("Writing to database {} {} {} {} {} {}".format(self.filename, domain, record_seen_by_web, record_expires, record_seen_by_isc, record_seen_by_you))
        sql = "insert or replace into domains (domain, seen_by_web,expires,seen_by_isc,seen_by_you) values (?,?,?,?,?)"
        with self.lock:
            cursor.execute(sql, (domain, record_seen_by_web, record_expires, record_seen_by_isc, record_seen_by_you))
            db.commit()
            self.stats.insert += 1
        return 1

    def delete_record(self, domain):
        log.info("Deleting record from database for {}".format(domain))
        db = sqlite3.connect(self.filename, timeout=15)
        cursor = db.cursor()
        with self.lock:
            cursor.execute("delete from domains where domain=?", (domain,))
            db.commit()
            self.stats.delete += 1
        return 1

    def get_record(self, domain):
        #Pass the timezone offset  hardcoded to utc for now
        #If record not found returns None,None,None,None
        #If record found rturns dates seen by web,expired,isc and you
        #If record is in database but domain registration expired it deletes the record and ignores it.
        timezone_offset = 0
        db = sqlite3.connect(self.filename, timeout=15)
        cursor = db.cursor()
        record = cursor.execute("select seen_by_web,expires, seen_by_isc, seen_by_you from domains where domain = ?" , (domain,) ).fetchone()
        if record:
            web,expires,isc,you = record
        else:
            self.stats.miss += 1
            log.info("No record in the database.  Returning None.")
            return (None,None,None,None)
        web = datetime.datetime.strptime(web, '%Y-%m-%d %H:%M:%S')
        expires = datetime.datetime.strptime(expires, '%Y-%m-%d %H:%M:%S')
        if expires < datetime.datetime.utcnow():
            log.info(f"Expired domain in database {domain} {expires}. Deleted")
            with self.lock:
                cursor.execute("delete from domains where domain=?", (domain,))
                db.commit()
                self.stats.delete += 1
            return (None,None,None,None)
        if isc != "LOCAL":
            isc = datetime.datetime.strptime(isc, '%Y-%m-%d %H:%M:%S')
        if you != "FIRST-CONTACT":
            you = datetime.datetime.strptime(you, '%Y-%m-%d %H:%M:%S')
        else:
            with self.lock:
                cursor.execute("update domains set seen_by_you=? where domain =?", ((datetime.datetime.utcnow()+datetime.timedelta(hours=timezone_offset)).strftime("%Y-%m-%d %H:%M:%S"), domain))
                db.commit()
        self.stats.hit += 1
        return (web,expires,isc,you)

    def process_update_file(self, update_file):
        """ Process csv in the format command, domain, web, expire, seen_by_isc """
        """ if command is + we add the record setting if it doesnt already exist"""
        """ if command is - we delete the record"""
        if not pathlib.Path(update_file).exists():
            log.info(f"The specified update file {update_file} does not exists.")
            return 0
        new_domains = open(update_file).readlines()
        num_recs = len(new_domains)
        db = sqlite3.connect(self.filename, timeout=15)
        cursor = db.cursor()    
        for pos,entry in enumerate(new_domains):
            if pos % 50 == 0:
                print("\r|{0:-<50}| {1:3.2f}%".format("X"*( 50 * pos//num_recs), 100*pos/num_recs),end="")
            command, domain, web, expires = entry.strip().split(",")
            domain = reduce_domain(domain)
            web = datetime.datetime.strptime(web, '%Y-%m-%d %H:%M:%S')
            expires = datetime.datetime.strptime(expires, '%Y-%m-%d %H:%M:%S')
            if command == "+":
                record = cursor.execute("select seen_by_web, expires, seen_by_isc, seen_by_you from domains where domain = ?" , (domain,) ).fetchone()
                if not record:
                    self.update_record(domain, web, expires,"LOCAL", "FIRST-CONTACT")
                    log.debug(f"Record added to database for domain {domain}")
                else:
                    log.debug(f"Record already exists skipped {domain}")
            elif command == "-":
                self.delete_record(domain)
                log.debug(f"Deleted record for {domain}")
        db.commit()
        print("\r|XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX| 100.00% FINISHED")
        return num_recs

    def update_database(self, latest_version, update_url):
        new_records_count = 0
        latest_major, latest_minor = map(int, str(latest_version).split("."))
        current_major, current_minor = map(int, str(self.version).split("."))
        dst_folder = pathlib.Path(self.filename).parent / "data" / f"{current_major}"
        if not dst_folder.exists():
            os.makedirs(dst_folder)
        log.info(f"Updating from {self.version} to {latest_version}")
        if latest_major > current_major:
            log.info("WARNING: Domain Stats database is a major revision behind. Database required rebuild.")
            raise Exception("WARNING: Domain Stats database is a major revision behind. Database required rebuild.")
        target_updates = range(current_minor+1, latest_minor+1 )
        for update in target_updates:
            version = f"{current_major}.{update}"
            log.info(f"Now applying update {version}")
            tgt_url = f"{update_url}/{current_major}/{update}.txt" 
            dst_path = pathlib.Path().cwd() / "data" / f"{current_major}" / f"{update}.txt"
            urllib.request.urlretrieve(tgt_url, str(dst_path))
            new_records_count += self.process_update_file(str(dst_path))
        self.version = latest_version
        self.lastupdate = datetime.datetime.utcnow()
        db = sqlite3.connect(self.filename, timeout=15)
        cursor = db.cursor()
        cursor.execute("update info set version=?, lastupdate=?",(self.version, self.lastupdate))
        db.commit()
        return latest_version, new_records_count


import rdap
import json
import datetime
import dateutil.parser
import dateutil.tz
import logging


def retrieve_data(action_name,eventlist):
    for entry in eventlist:
        if entry.get("eventAction") == action_name:
            return entry.get("eventDate")
    return None

def get_domain_record(domain):
    client = rdap.client.RdapClient()
    client.url = "https://www.rdap.net"
    client.timeout = 5
    try:
        resp = client.get_domain(domain).data
    except Exception as e:
        return "ERROR","ERROR", str(e)
    reg = retrieve_data( 'registration', resp.get('events'))
    reg = dateutil.parser.isoparse(reg)
    reg.replace(tzinfo=datetime.timezone.utc)
    exp = retrieve_data( 'expiration', resp.get('events'))
    exp = dateutil.parser.isoparse(exp)
    exp.replace(tzinfo=datetime.timezone.utc)
    return reg,exp,""


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="domain_stats", 
    version="0.0.6",
    author="MarkBaggett",
    author_email="lo127001@gmail.com",
    description="Malicious Domain Detection base on domain creation and first contact",
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts = ['domain_stats/domain_stats', "domain_stats/domain_stats_db_admin"],
    url="https://github.com/markbaggett/domain_stats",
    packages=setuptools.find_packages(),
    install_requires = ['requests','pyyaml', 'rdap', 'python-dateutil'],
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from collections import UserDict
import yaml

class Config(UserDict):

    def __init__(self, filename=""):
        super().__init__()
        self.filename = filename
        self.load_config(filename)

    def load_config(self, filename = None):
        filename = filename or self.filename
        with open(filename) as fh:
            self.data = yaml.safe_load(fh.read())

    def save_config(self, filename = None):
        filename = filename or self.filename
        with open(filename, 'w') as fp:
            yaml.dump(self.data, fp, default_flow_style=False)
        